"""
This file is part of pyS5p

https://github.com/rmvanhees/pys5p.git

The class S5Pmsm read HDF5 measurement data including its attributes and
dimensions. Initialization:

  S5Pmsm attribute | hdf5 dataset           | Numpy array
  -------------------------------------------------------------------------
  name             | h5_dset.name           | 'value'
  value            | h5_dset.value['value'] | np.squeeze(data)
                   | or h5_dset.value       |
  error            | h5_dset.value['error'] | None
                   | or None                |
  coords           | h5_dset.dims           | [[['time',] 'row',] 'column']
  units            | attrs['units']         | None
  long_name        | attrs['long_name']     | ''
  fillvalue        | h5_dset.fillvalue      | None
  coverage         | None                   | None

Limited to 3 dimensions

Copyright (c) 2017-2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from collections import namedtuple
from copy import deepcopy
from pathlib import PurePath

from h5py import Dataset
import numpy as np

from .biweight import biweight

# - local functions --------------------------------
def pad_rows(arr1, arr2):
    """
    Pad the array with the least numer of rows with NaN's
    """
    if arr2.ndim == 2:
        if arr1.shape[0] < arr2.shape[0]:
            buff = arr1.copy()
            arr1 = np.full_like(arr2, np.nan)
            arr1[0:buff.shape[0], :] = buff
        elif arr1.shape[0] > arr2.shape[0]:
            buff = arr2.copy()
            arr2 = np.full_like(arr1, np.nan)
            arr2[0:buff.shape[0], :] = buff
    else:
        if arr1.shape[1] < arr2.shape[1]:
            buff = arr1.copy()
            arr1 = np.full_like(arr2, np.nan)
            arr1[:, 0:buff.shape[1], :] = buff
        elif arr1.shape[1] > arr2.shape[1]:
            buff = arr2.copy()
            arr2 = np.full_like(arr1, np.nan)
            arr2[:, 0:buff.shape[1], :] = buff

    return (arr1, arr2)


# - class definition -------------------------------
class S5Pmsm():
    """
    Definition of class S5Pmsm which contains the data of a HDF5 dataset,
    including its attributes (CF convension)

    Attributes
    ----------
    name : string
       Name of the original HDF5 dataset
    value : ndarray
       Dataset values
    error : ndarray, optional
       Dataset uncertainties or error estimates
    coords : OrderDict
       Dataset coordinates (HDF5/netCDF4 dimensions)
    coverage : string, optional
       Dataset coverage start and end
    units : string
       Data units
    long_name : string
       Dataset long name
    fillvalue : float
       Value of undefined or missings data values

    Methods
    -------
    copy()
       Return a deep copy of the current object.
    set_coords(coords_data, coords_name=None)
       Set coordinates of data.
    set_coverage(coverage, force=False)
       Set the coverage attribute, as (coverageStart, coverageEnd),
    set_units(units, force=False)
       Set the units attribute.
    set_fillvalue()
       Set fillvalue of floating-point data.
    set_long_name(name, force=False)
       Set the long_name attribute.
    fill_as_nan()
       Replace missing floating-point data with NaN's.
    sort(axis=0)
       Sort dataset according a coordinate axis.
    concatenate(msm, axis=0)
       Concatenate the data of a S5Pmsm to the current S5Pmsm object.
    nanpercentile(vperc, data_sel=None, axis=0, keepdims=False)
       Returns percentile(s) of the data in the S5Pmsm object.
    biweight(data_sel=None, axis=0, keepdims=False)
       Returns biweight median of the data in the S5Pmsm object.
    nanmedian(data_sel=None, axis=0, keepdims=False)
       Returns S5Pmsm object containing median & standard deviation of the
        original data.
    nanmean(data_sel=None, axis=0, keepdims=False)
       Returns S5Pmsm object containing mean & standard deviation of the
        original data.
    transpose()
       Transpose data and coordinates of an S5Pmsm object.

    Notes
    -----

    Examples
    --------
    """
    def __init__(self, dset, data_sel=None, datapoint=False):
        """
        Read measurement data from a Tropomi OCAL, ICM, of L1B product

        Parameters
        ----------
        dset      :  h5py.Dataset or ndarray
           h5py dataset from which the data is read, data is used to
           initalize S5Pmsm object
        data_sel  :  numpy slice
           a numpy slice generated for example numpy.s_
        datapoint :  boolean
           to indicate that the dataset is a compound of type datapoint

        Returns
        -------
        numpy structure with dataset data and attributes, including data,
        fillvalue, coordinates, units, ...

        """
        # initialize object
        self.name = 'value'
        self.value = None
        self.error = None
        self.coords = None
        self.coverage = None
        self.units = None
        self.long_name = ''
        self.fillvalue = None

        if isinstance(dset, Dataset):
            self.__from_h5_dset(dset, data_sel, datapoint)
        else:
            self.__from_ndarray(dset, data_sel)

    def __repr__(self) -> str:
        res = []
        for attr in self.__dict__:
            if attr.startswith('__'):
                continue
            if isinstance(self.__dict__[attr], np.ndarray):
                value = self.__dict__[attr].shape
            else:
                value = self.__dict__[attr]
            res.append('{}: {}'.format(attr, value))
        return '\n'.join(res)

    def __from_h5_dset(self, h5_dset, data_sel, datapoint):
        """
        initialize S5Pmsm object from h5py dataset
        """
        self.name = PurePath(h5_dset.name).name

        # copy dataset values (and error) to object
        if data_sel is None:
            if datapoint:
                self.value = h5_dset['value']
                self.error = h5_dset['error']
            else:
                self.value = h5_dset[...]
        else:
            # we need to keep all dimensions to get the dimensions
            # of the output data right
            if datapoint:
                self.value = h5_dset['value'][data_sel]
                self.error = h5_dset['error'][data_sel]
                if isinstance(data_sel, tuple):
                    for ii, elmnt in enumerate(data_sel):
                        if isinstance(elmnt, (int, np.int64)):
                            self.value = np.expand_dims(self.value, axis=ii)
                            self.error = np.expand_dims(self.error, axis=ii)
            else:
                self.value = h5_dset[data_sel]
                if isinstance(data_sel, tuple):
                    for ii, elmnt in enumerate(data_sel):
                        if isinstance(elmnt, (int, np.int64)):
                            self.value = np.expand_dims(self.value, axis=ii)

        # set default dimension names
        if h5_dset.ndim == 1:
            keys_default = ['column']
        elif h5_dset.ndim == 2:
            keys_default = ['row', 'column']
        elif h5_dset.ndim == 3:
            keys_default = ['time', 'row', 'column']
        else:
            raise ValueError('not implemented for ndim > 3')

        # copy all dimensions with size longer then 1
        keys = []
        dims = []
        for ii in range(h5_dset.ndim):
            if self.value.shape[ii] == 1:
                continue

            if len(h5_dset.dims[ii]) != 1:   # bug in some KMNI HDF5 files
                keys.append(keys_default[ii])
                dims.append(np.arange(self.value.shape[ii]))
            elif self.value.shape[ii] == h5_dset.shape[ii]:
                buff = PurePath(h5_dset.dims[ii][0].name).name
                if len(buff.split()) > 1:
                    buff = buff.split()[0]
                keys.append(buff)
                if h5_dset.dims[ii][0][:].size == h5_dset.shape[ii]:
                    buff = h5_dset.dims[ii][0][:]
                    if np.all(buff == 0):
                        buff = np.arange(buff.size)
                else:                          # bug in some KMNI HDF5 files
                    buff = np.arange(h5_dset.shape[ii])
                dims.append(buff)
            else:
                buff = PurePath(h5_dset.dims[ii][0].name).name
                if len(buff.split()) > 1:
                    buff = buff.split()[0]
                keys.append(buff)
                if h5_dset.dims[ii][0][:].size == h5_dset.shape[ii]:
                    buff = h5_dset.dims[ii][0][:]
                    if np.all(buff == 0):
                        buff = np.arange(buff.size)
                else:                          # bug in some KMNI HDF5 files
                    buff = np.arange(h5_dset.shape[ii])

                if isinstance (data_sel, slice):
                    dims.append(buff[data_sel])
                elif len(data_sel) == h5_dset.ndim:
                    dims.append(buff[data_sel[ii]])
                elif not isinstance(data_sel, tuple):
                    dims.append(buff[data_sel])
                elif ii > len(data_sel):
                    dims.append(buff[data_sel[-1]])
                else:
                    dims.append(buff[data_sel[ii]])

        # add dimensions as a namedtuple
        coords_namedtuple = namedtuple('Coords', keys)
        self.coords = coords_namedtuple._make(dims)

        # remove all dimensions with size equal 1 from value (and error)
        self.value = np.squeeze(self.value)
        if datapoint:
            self.error = np.squeeze(self.error)

        # copy FillValue (same for value/error in a datapoint)
        if datapoint:
            self.fillvalue = h5_dset.fillvalue[0]
        else:
            self.fillvalue = h5_dset.fillvalue

        # copy its units
        if 'units' in h5_dset.attrs:
            if isinstance(h5_dset.attrs['units'], np.ndarray):
                if h5_dset.attrs['units'].size == 1:
                    self.units = h5_dset.attrs['units'][0]
                    if isinstance(self.units, bytes):
                        self.units = self.units.decode('ascii')
                else:
                    self.units = h5_dset.attrs['units']
                    if isinstance(self.units[0], bytes):
                        self.units = self.units.astype(str)
            else:
                self.units = h5_dset.attrs['units']
                if isinstance(self.units, bytes):
                    self.units = self.units.decode('ascii')

        # copy its long_name
        if 'long_name' in h5_dset.attrs:
            if isinstance(h5_dset.attrs['long_name'], bytes):
                self.long_name = h5_dset.attrs['long_name'].decode('ascii')
            else:
                self.long_name = h5_dset.attrs['long_name']

    def __from_ndarray(self, data, data_sel):
        """
        initialize S5Pmsm object from a ndarray
        """
        # copy dataset values (and error) to object
        if data_sel is None:
            self.value = np.squeeze(data)
        else:
            self.value = np.squeeze(data[data_sel])

        # define coordinates
        dims = [np.arange(sz) for sz in self.value.shape]
        try:
            self.set_coords(dims, coords_name=None)
        except Exception as exc:
            raise RuntimeError('failed to set the coordinates') from exc

    def copy(self):
        """
        return a deep copy of the current object
        """
        return deepcopy(self)

    def set_coords(self, coords_data, coords_name=None):
        """
        Set coordinates of data

        Parameters
        ----------
        coords_data  :  list of ndarrays
          list with coordinates data for each dimension
        coords_name  :  list of strings
          list with the names of each dimension
        """
        if coords_name is None:
            if len(coords_data) == 1:
                keys = ['column']
            elif len(coords_data) == 2:
                keys = ['row', 'column']
            elif len(coords_data) == 3:
                keys = ['time', 'row', 'column']
            else:
                raise ValueError('not implemented for ndim > 3')
        else:
            if isinstance(coords_name, str):
                keys = [coords_name]
            else:
                keys = coords_name

        # add dimensions as a namedtuple
        coords_namedtuple = namedtuple('Coords', keys)
        self.coords = coords_namedtuple._make(coords_data)

    def set_coverage(self, coverage, force=False):
        """
        Set the coverage attribute, as (coverageStart, coverageEnd)
        Both elements are expected to be datatime objects.
        Overwrite when force is true
        """
        if self.coverage is None or force:
            self.coverage = coverage

    def set_units(self, units, force=False):
        """
        Set the units attribute, overwrite when force is true
        """
        if self.units is None or force:
            self.units = units

    def set_fillvalue(self):
        """
        Set fillvalue to KNMI undefined
        """
        if np.issubdtype(self.value.dtype, np.floating):
            if self.fillvalue is None or self.fillvalue == 0.:
                self.fillvalue = float.fromhex('0x1.ep+122')

    def set_long_name(self, name, force=False):
        """
        Set the long_name attribute, overwrite when force is true
        """
        if force or not self.long_name:
            self.long_name = name

    def fill_as_nan(self):
        """
        Replace fillvalues in data with NaN's

        Works only on datasets with HDF5 datatype 'float' or 'datapoints'
        """
        if self.fillvalue == float.fromhex('0x1.ep+122'):
            self.value[(self.value == self.fillvalue)] = np.nan
            if self.error is not None:
                self.error[(self.error == self.fillvalue)] = np.nan

    def sort(self, axis=0):
        """
        Sort data and its coordinate along a given axis

        Parameters
        ----------
        axis  : int, optional
           axis for which the array will be sorted. Default is 0.
        """
        if not isinstance(axis, int):
            raise TypeError('axis not an integer')
        if not 0 <= axis < self.value.ndim:
            raise ValueError('axis out-of-range')

        indx = np.argsort(self.coords[axis][:])
        self.coords[axis][:] = self.coords[axis][indx]

        if axis == 0:
            self.value = self.value[indx, ...]
            if self.error is not None:
                if isinstance(self.error, list):
                    self.error = (self.error[0][indx, ...],
                                  self.error[1][indx, ...])
                else:
                    self.error = self.error[indx, ...]
        elif axis == 1:
            self.value = self.value[:, indx, ...]
            if self.error is not None:
                if isinstance(self.error, list):
                    self.error = (self.error[0][:, indx, :],
                                  self.error[1][:, indx, :])
                else:
                    self.error = self.error[:, indx, :]
        elif axis == 2:
            self.value = self.value[:, :, indx]
            if self.error is not None:
                if isinstance(self.error, list):
                    self.error = (self.error[0][:, :, indx],
                                  self.error[1][:, :, indx])
                else:
                    self.error = self.error[:, :, indx]
        else:
            raise ValueError("S5Pmsm: implemented for ndim <= 3")

    def concatenate(self, msm, axis=0):
        """
        Concatenate two measurement datasets, the current with another.

        Parameters
        ----------
        msm   :  pys5p.S5Pmsm
           an S5Pmsm object
        axis  : int, optional
           The axis for which the array will be joined. Default is 0.

        Returns
        -------
        The data of the new dataset is concatenated to the existing data along
        an existing axis. The affected coordinate is also extended.

        Note:
         - The arrays must have the same shape, except in the dimension
        corresponding to axis (the first, by default).
        """
        if self.name != PurePath(msm.name).name:
            raise TypeError('combining dataset with different name')

        if self.error is None and msm.error is None:
            datapoint = False
        elif self.error is not None and msm.error is not None:
            datapoint = True
        else:
            raise RuntimeError("S5Pmsm: combining non-datapoint and datapoint")

        # all but the last 2 dimensions have to be equal
        if self.value.shape[:-2] != msm.value.shape[:-2]:
            raise TypeError('all but the last 2 dimensions should be equal')

        if axis == 0:
            self.value = np.concatenate((self.value, msm.value), axis=axis)
            if datapoint:
                self.error = np.concatenate((self.error, msm.error),
                                            axis=axis)
        elif axis == 1:
            if self.value.shape[0] == msm.value.shape[0]:
                self.value = np.concatenate((self.value, msm.value), axis=axis)
                if datapoint:
                    self.error = np.concatenate((self.error, msm.error),
                                                axis=axis)
            else:
                self.value = np.concatenate(pad_rows(self.value, msm.value),
                                            axis=axis)
                if datapoint:
                    self.error = np.concatenate(
                        pad_rows(self.error, msm.error), axis=axis)
        elif axis == 2:
            if self.value.shape[1] == msm.value.shape[1]:
                self.value = np.concatenate((self.value, msm.value), axis=axis)
                if datapoint:
                    self.error = np.concatenate((self.error, msm.error),
                                                axis=axis)
            else:
                self.value = np.concatenate(pad_rows(self.value, msm.value),
                                            axis=axis)
                if datapoint:
                    self.error = np.concatenate(
                        pad_rows(self.error, msm.error), axis=axis)
        else:
            raise ValueError("S5Pmsm: implemented for ndim <= 3")

        # now extent coordinate of the fastest axis
        key = self.coords._fields[axis]
        if msm.coords[axis][0] == 0:
            dims = np.concatenate((self.coords[axis],
                                   len(self.coords[axis]) + msm.coords[axis]))
        else:
            dims = np.concatenate((self.coords[axis], msm.coords[axis]))
        self.coords = self.coords._replace(**{key: dims})
        return self

    def nanpercentile(self, vperc, data_sel=None, axis=0, keepdims=False):
        """
        Returns percentile(s) of the data in the S5Pmsm

        Parameters
        ----------
        vperc       :  list
           range to normalize luminance data between percentiles min and max of
           array data.
        data_sel  :  numpy slice
           A numpy slice generated for example numpy.s_. Can be used to skip
           the first and/or last frame
        axis      : int, optional
           Axis or axes along which the medians are computed. Default is 0.
        keepdims  : bool, optional
           If this is set to True, the axes which are reduced are left in the
           result as dimensions with size one. With this option, the result
           will broadcast correctly against the original arr.

        Returns
        -------
        S5Pmsm object with the original data replaced by the percentiles along
        one of the axis, see below. The coordinates are adjusted, accordingly.

        You should atleast supply one percentile and atmost three.
         vperc is instance 'int' or len(vperc) == 1:
             'value' is replaced by its (nan-)percentile vperc
             'error' is unchanged
         len(vperc) == 2:
             'vperc' is sorted
             'value' is replaced by its (nan-)median
             'error' is replaced by percentile('value', (vperc[0], vperc[1]))
         len(vperc) == 3:
             'vperc' is sorted
             'value' is replaced by percentile('value', vperc[1])
             'error' is replaced by percentile('value', (vperc[0], vperc[2]))
        """
        if isinstance(axis, int):
            axis = (axis,)

        if isinstance(vperc, int):
            vperc = (vperc,)
        else:
            if len(vperc) == 2:
                vperc += (50,)
            # make sure that the values are sorted
            vperc = tuple(sorted(vperc))

        if len(vperc) != 1 and len(vperc) != 3:
            raise TypeError('dimension vperc must be 1 or 3')

        if data_sel is None:
            if self.value.size <= 1 or self.value.ndim <= max(axis):
                return self
            perc = np.nanpercentile(self.value, vperc,
                                    axis=axis, keepdims=keepdims)
        else:
            if self.value[data_sel].size <= 1 \
               or self.value[data_sel].ndim <= max(axis):
                return self
            perc = np.nanpercentile(self.value[data_sel], vperc,
                                    axis=axis, keepdims=keepdims)
        if len(vperc) == 3:
            self.value = perc[1, ...]
            self.error = [perc[0, ...], perc[2, ...]]
        else:
            self.value = perc[0, ...]

        # adjust the coordinates
        if keepdims:
            key = self.coords._fields[axis]
            if self.coords[axis][0] == 0:
                dims = [0]
            else:
                dims = np.median(self.coords[axis], keepdims=keepdims)
            self.coords = self.coords._replace(**{key: dims})
        else:
            keys = []
            dims = []
            for ii in range(self.value.ndim+len(axis)):
                if ii not in axis:
                    keys.append(self.coords._fields[ii])
                    dims.append(self.coords[ii][:])
            coords_namedtuple = namedtuple('Coords', keys)
            self.coords = coords_namedtuple._make(dims)

        return self

    def biweight(self, data_sel=None, axis=0, keepdims=False):
        """
        Returns biweight median of the data in the S5Pmsm

        Parameters
        ----------
        data_sel  :  numpy slice
           A numpy slice generated for example numpy.s_. Can be used to skip
           the first and/or last frame
        axis  : int, optional
           Axis or axes along which the medians are computed. Default is 0.
        keepdims  : bool, optional
           If this is set to True, the axes which are reduced are left in the
           result as dimensions with size one. With this option, the result
           will broadcast correctly against the original arr.

        Returns
        -------
        S5Pmsm object with its data (value & error) replaced by its biweight
        medians along one axis. The coordinates are adjusted, accordingly.
        """
        if data_sel is None:
            if self.error is not None:
                self.value = biweight(self.value, axis=axis)
                self.error = biweight(self.error, axis=axis)
            else:
                (self.value, self.error) = biweight(self.value,
                                                    axis=axis, spread=True)
        else:
            if self.error is not None:
                self.value = biweight(self.value[data_sel], axis=axis)
                self.error = biweight(self.error[data_sel], axis=axis)
            else:
                (self.value, self.error) = biweight(self.value[data_sel],
                                                    axis=axis, spread=True)
        if keepdims:
            self.value = np.expand_dims(self.value, axis=axis)
            self.error = np.expand_dims(self.error, axis=axis)

        # adjust the coordinates
        if keepdims:
            key = self.coords._fields[axis]
            if self.coords[axis][0] == 0:
                dims = [0]
            else:
                dims = np.median(self.coords[axis], keepdims=keepdims)
            self.coords = self.coords._replace(**{key: dims})
        else:
            keys = []
            dims = []
            for ii in range(self.value.ndim+1):
                if ii != axis:
                    keys.append(self.coords._fields[ii])
                    dims.append(self.coords[ii][:])
            coords_namedtuple = namedtuple('Coords', keys)
            self.coords = coords_namedtuple._make(dims)

        return self

    def nanmedian(self, data_sel=None, axis=0, keepdims=False):
        """
        Returns S5Pmsm object containing median & standard deviation of the
        original data

        Parameters
        ----------
        data_sel  :  numpy slice, optional
           A numpy slice generated for example numpy.s_. Can be used to skip
           the first and/or last frame
        axis      : int, optional
           Axis or axes along which the medians are computed. Default is 0.
        keepdims  : bool, optional
           If this is set to True, the axes which are reduced are left in the
           result as dimensions with size one. With this option, the result
           will broadcast correctly against the original arr.

        Returns
        -------
        S5Pmsm object with its data (value & error) replaced by its nanmedian
        and standard deviation along one axis.
        The coordinates are adjusted, accordingly.
        """
        if data_sel is None:
            if self.error is not None:
                self.error = np.nanmedian(self.error,
                                          axis=axis, keepdims=keepdims)
            else:
                self.error = np.nanstd(self.value, ddof=1,
                                       axis=axis, keepdims=keepdims)
            self.value = np.nanmedian(self.value, axis=axis, keepdims=keepdims)
        else:
            if self.error is not None:
                self.error = np.nanmedian(self.error[data_sel],
                                          axis=axis, keepdims=keepdims)
            else:
                self.error = np.nanstd(self.value[data_sel], ddof=1,
                                       axis=axis, keepdims=keepdims)
            self.value = np.nanmedian(self.value[data_sel],
                                      axis=axis, keepdims=keepdims)

        # adjust the coordinates
        if keepdims:
            key = self.coords._fields[axis]
            if self.coords[axis][0] == 0:
                dims = [0]
            else:
                dims = np.median(self.coords[axis], keepdims=keepdims)
            self.coords = self.coords._replace(**{key: dims})
        else:
            keys = []
            dims = []
            for ii in range(self.value.ndim+1):
                if ii != axis:
                    keys.append(self.coords._fields[ii])
                    dims.append(self.coords[ii][:])
            coords_namedtuple = namedtuple('Coords', keys)
            self.coords = coords_namedtuple._make(dims)

        return self

    def nanmean(self, data_sel=None, axis=0, keepdims=False):
        """
        Returns S5Pmsm object containing mean & standard deviation of the
        original data

        Parameters
        ----------
        data_sel  :  numpy slice, optional
           A numpy slice generated for example numpy.s_. Can be used to skip
           the first and/or last frame
        axis      : int, optional
           Axis or axes along which the mean are computed. Default is 0.
        keepdims  : bool, optional
           If this is set to True, the axes which are reduced are left in the
           result as dimensions with size one. With this option, the result
           will broadcast correctly against the original arr.

        Returns
        -------
        S5Pmsm object with its data (value & error) replaced by its nanmean and
        standard deviation along one axis.
        The coordinates are adjusted, accordingly.
        """
        if data_sel is None:
            if self.error is not None:
                self.error = np.nanmean(self.error,
                                        axis=axis, keepdims=keepdims)
            else:
                self.error = np.nanstd(self.value, ddof=1,
                                       axis=axis, keepdims=keepdims)
            self.value = np.nanmean(self.value, axis=axis, keepdims=keepdims)
        else:
            if self.error is not None:
                self.error = np.nanmean(self.error[data_sel],
                                        axis=axis, keepdims=keepdims)
            else:
                self.error = np.nanstd(self.value[data_sel], ddof=1,
                                       axis=axis, keepdims=keepdims)
            self.value = np.nanmean(self.value[data_sel],
                                    axis=axis, keepdims=keepdims)

        # adjust the coordinates
        if keepdims:
            key = self.coords._fields[axis]
            if self.coords[axis][0] == 0:
                dims = [0]
            else:
                dims = np.mean(self.coords[axis], keepdims=keepdims)
            self.coords = self.coords._replace(**{key: dims})
        else:
            keys = []
            dims = []
            for ii in range(self.value.ndim+1):
                if ii != axis:
                    keys.append(self.coords._fields[ii])
                    dims.append(self.coords[ii][:])
            coords_namedtuple = namedtuple('Coords', keys)
            self.coords = coords_namedtuple._make(dims)

        return self

    def transpose(self):
        """
        Transpose data and coordinates of an S5Pmsm object
        """
        if self.value.ndim <= 1:
            return self

        if self.error is not None:
            self.error = np.transpose(self.error)
        self.value = np.transpose(self.value)

        keys = []
        dims = []
        for ii in range(self.value.ndim):
            keys.append(self.coords._fields[ii])
            dims.append(self.coords[ii][:])
        tmp = keys[1]
        keys[1] = keys[0]
        keys[0] = tmp
        tmp = dims[1]
        dims[1] = dims[0]
        dims[0] = tmp
        coords_namedtuple = namedtuple('Coords', keys)
        self.coords = coords_namedtuple._make(dims)

        return self



"""
This file is part of pyS5p

https://github.com/rmvanhees/pys5p.git

Performs unit-tests on S5Pplot methods: draw_signal, draw_quality,
   draw_cmp_images, draw_trend1d and draw_lines

Copyright (c) 2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import argparse
from pathlib import Path

import numpy as np

from pys5p.lib.plotlib import FIGinfo
from pys5p.l1b_io import L1BioRAD
from pys5p.lv2_io import LV2io
from pys5p.s5p_geoplot import S5Pgeoplot


# --------------------------------------------------
def run_l1b_tests(plot, l1b_product):
    """
    Run unit tests using an L1B product
    """
    print('Run unit tests on S5Pgeoplot using L1BioRAD')

    with L1BioRAD(l1b_product) as l1b:
        orbit = l1b.get_orbit()
        print('orbit: ', orbit)

        _ = l1b.select()

        seq = l1b.sequence()
        geo_sat = l1b.get_geo_data(
            geo_dset='satellite_latitude,satellite_longitude')

        geo = l1b.get_geo_data()

        l1b_data = l1b.get_msm_data('radiance', fill_as_nan=True)

    mask = (seq['icid'] == 12) | (seq['icid'] == 14)
    for key in geo_sat:
        print(key, geo_sat[key][mask].min(), geo_sat[key][mask].max())
    fig_info_in = FIGinfo()
    fig_info_in.add('ICIDs', [12, 14])
    plot.draw_geo_subsat(geo_sat['satellite_longitude'][mask],
                         geo_sat['satellite_latitude'][mask],
                         title='S5Pgeoplot::draw_geo_subsat',
                         sub_title='orbit={}'.format(orbit),
                         fig_info=fig_info_in.copy())

    plot.draw_geo_tiles(geo['longitude'][mask, :],
                        geo['latitude'][mask, :],
                        title='S5Pgeoplot::draw_geo_tiles',
                        sub_title='orbit={}'.format(orbit),
                        fig_info=fig_info_in.copy())

    plot.draw_geo_tiles(geo['longitude'][mask, :], geo['latitude'][mask, :],
                        sequence=seq['sequence'][mask],
                        title='S5Pgeoplot::draw_geo_tiles',
                        sub_title='orbit={}'.format(orbit),
                        fig_info=fig_info_in.copy())

    plot.draw_geo_msm(geo['longitude'][mask, :], geo['latitude'][mask, :],
                      np.nanmean(l1b_data, axis=2)[mask, :],
                      title='S5Pgeoplot::draw_geo_msm',
                      sub_title='orbit={}'.format(orbit),
                      fig_info=fig_info_in.copy())


def run_lv2_tests(plot, lv2_product):
    """
    Run unit tests using an LV2 product
    """
    print('Run unit tests on S5Pgeoplot using LV2io')

    with LV2io(lv2_product) as lv2:
        orbit = lv2.get_orbit()
        print('orbit: ', orbit)

        extent = [-180, 180, -30, 30]
        data_sel, geo = lv2.get_geo_bounds(extent=extent)
        print('data_sel: ', data_sel)
        for key in geo:
            print(key, geo[key].shape)

        if lv2.science_product:
            msm_name = 'co_column'
        else:
            msm_name = 'carbonmonoxide_total_column'
        lv2_msm = lv2.get_data_as_s5pmsm(msm_name, data_sel=data_sel,
                                         fill_as_nan=True)

    print(lv2_msm.long_name)
    plot.draw_geo_msm(geo['longitude'], geo['latitude'], lv2_msm,
                      title=lv2_msm.long_name,
                      sub_title='orbit={}'.format(orbit))


# --------------------------------------------------
def main():
    """
    main function
    """
    # parse command-line parameters
    parser = argparse.ArgumentParser(
        description='{}: run units-test on class S5Pgeoplot'.format(
            Path(__file__).name))
    parser.add_argument('--l1b_file', type=str, default=None,
                        help='Use S5P_L1B_RAD_PRODUCT to run unit-test on ...')
    parser.add_argument('--lv2_file', type=str, default=None,
                        help='Use S5P_LV2_PRODUCT to run unit-test on ...')

    args = parser.parse_args()

    if args.l1b_file is None and args.lv2_file is None:
        parser.print_help()
        return

    plot = S5Pgeoplot('unit_test_s5p_geoplot.pdf')
    if args.l1b_file is not None:
        if not Path(args.l1b_file).is_file():
            raise FileNotFoundError('S5P_L1B_RAD_PRODUCT not found')
        run_l1b_tests(plot, args.l1b_file)

    if args.lv2_file is not None:
        if not Path(args.lv2_file).is_file():
            raise FileNotFoundError('S5P_LV2_PRODUCT not found')
        run_lv2_tests(plot, args.lv2_file)

    # close figure
    plot.close()


# - main code --------------------------------------
if __name__ == '__main__':
    main()



# This file is part of pys5p
#
# https://github.com/rmvanhees/pys5p.git
#
# Copyright (c) 2017-2020 SRON - Netherlands Institute for Space Research
#   All Rights Reserved
#
# License:  BSD-3-Clause

""" The pys5p package contains software to read S5p Tropomi L1B products.
    And contains plotting routines to display your data beautifully."""

from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    pass



"""
This file is part of pyS5p

https://github.com/rmvanhees/pys5p.git

This module contains the helper functions for the class S5Pplot

Copyright (c) 2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from collections import OrderedDict
from copy import deepcopy
from datetime import datetime

import matplotlib as mpl
import numpy as np
import xarray as xr


# set the colormap and centre the colorbar
class MidpointNormalize(mpl.colors.Normalize):
    """
    Normalise the colorbar so that diverging bars work there way either side
    from a prescribed midpoint value)

    e.g. im = ax1.imshow(array, norm=MidpointNormalize(midpoint=0.,
                                                       vmin=-100, vmax=100))
    """
    def __init__(self, vmin=None, vmax=None, midpoint=None, clip=False):
        self.midpoint = midpoint
        mpl.colors.Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        # I'm ignoring masked values and all kinds of edge cases to make a
        # simple example...
        xx, yy = [self.vmin, self.midpoint, self.vmax], [0, 0.5, 1]
        return np.ma.masked_array(np.interp(value, xx, yy), np.isnan(value))


# define a class to handle the figure information
class FIGinfo:
    """
    The figure information constists of key, value combinations which are
    to be displayed 'above' or 'right' from the main image.

    Attributes
    ----------
    location : string
       Location to draw the fig_info box: 'above' (default), 'right', 'none'
    fig_info : OrderedDict
       Dictionary holding the information for the fig_info box

    Methods
    -------
    add(key, value, fmt='{}')
       Extent fig_info with a new line.
    as_str()
       Return figure information as one long string.
    copy()
       Return a deep copy of the current object.
    set_location(loc)
       Set the location of the fig_info box.

    Notes
    -----
    The box with the figure information can only hold a limited number of keys:
      'above' :  The figure information is displayed in a small box. This box
                 grows with the number of lines and will overlap with the main
                 image or its colorbar at about 7+ entries.
      'right' :  Depending on the aspect ratio of the main image, the number of
                 lines are limited to 70 (aspect-ratio=1), 50 (aspect-ratio=2),
                 or 40 (aspect-ratio > 2).
    """
    def __init__(self, loc='above', info_dict=None) -> None:
        self.fig_info = OrderedDict() if info_dict is None else info_dict
        self.set_location(loc)

    def __bool__(self) -> bool:
        return bool(self.fig_info)

    def __len__(self) -> int:
        return len(self.fig_info)

    def copy(self):
        """
        Return a deep copy of the current object
        """
        return deepcopy(self)

    def set_location(self, loc: str) -> None:
        """
        Set the location of the fig_info box

        Parameters
        ----------
        loc : str
          Location of the fig_info box
        """
        if loc not in ('above', 'right', 'none'):
            raise KeyError('location should be: above, right or none')

        self.location = loc

    def add(self, key: str, value, fmt='{}') -> None:
        """
        Extent fig_info with a new line

        Parameters
        ----------
        key : str
          Name of the fig_info key.
        value : Any python variable
          Value of the fig_info key. A tuple will be formatted as *value.
        fmt : str
          Convert value to a string, using the string format method.
          Default: '{}'.
        """
        if isinstance(value, tuple):
            self.fig_info.update({key: fmt.format(*value)})
        else:
            self.fig_info.update({key: fmt.format(value)})

    def as_str(self) -> str:
        """
        Return figure information as one long string
        """
        info_str = ""
        for key in self.fig_info:
            info_str += "{} : {}\n".format(key, self.fig_info[key])

        # add timestamp
        info_str += 'created : {}'.format(
            datetime.utcnow().isoformat(timespec='seconds'))

        return info_str


def blank_legend_key():
    """
    Show only text in matplotlib legenda, no key
    """
    return mpl.patches.Rectangle((0, 0), 0, 0, fill=False,
                                 edgecolor='none', visible=False)


def check_data2d(method: str, data) -> None:
    """
    Make sure that the input data is 2-dimensional and contains valid data
    """
    # data is stored in a numpy.ndarray
    if isinstance(data, np.ndarray):
        if data.ndim != 2:
            raise ValueError('data must be two dimensional')
        if np.all(np.isnan(data)):
            raise ValueError('data must contain valid data')
        return

    # data must be stored in a xarray.DataArray object
    if not isinstance(data, xr.DataArray):
        raise TypeError('data not a numpy.ndarray or xarray.DataArray')

    if data.values.ndim != 2:
        raise ValueError('data must be two dimensional')
    if np.all(np.isnan(data.values)):
        raise ValueError('data must contain valid data')
    if method == 'ratio_unc':
        if data.error.ndim != 2:
            raise ValueError('data must be two dimensional')
        if np.all(np.isnan(data.error)):
            raise ValueError('data must contain valid error estimates')


def data_for_trend2d(data_in, coords=None, delta_time=None):
    """
    Prepare data to display the averaged column/row as function of time

    Parameters
    ----------
    data :  numpy.ndarray or xarray.DataArray
       Object holding measurement data and attributes
    coords :  dictionary, optional
       Dictionary with the names and data of each dimension, where the keys of
       the  dictionary are the names of the dimensions.
       Required when data is a numpy array (otherwise ignored)
    delta_time : int, optional
       Expected resolution of the time-axis.
       Defined in whole seconds or number of orbit revolutions.
    missing_data: 'NaN' or 'previous'
       Default: NaN

    Returns
    -------
    xarray.DataArray
    """
    try:
        check_data2d('data', data_in)
    except Exception as exc:
        raise RuntimeError('invalid input-data provided') from exc

    if isinstance(data_in, np.ndarray):
        if coords is None:
            raise KeyError('coords needs to be provided when data is a ndarray')
        msm = data_to_xr(data_in, coords=coords)
    else:
        msm = data_in

    # determine time-axis
    (ylabel, xlabel) = msm.coords._fields
    if ylabel in ('orbit', 'time'):
        time_axis = 0
        time_data = msm.coords[0]
    elif xlabel in ('orbit', 'time'):
        time_axis = 1
        time_data = msm.coords[1]
    else:
        raise RuntimeError('can not find the time-axis')

    if not np.issubdtype(time_data.dtype, np.integer):
        raise ValueError('time-coordinate not of type integer')
    if not np.all(time_data[1:] > time_data[:-1]):
        indx = (time_data[1:] <= time_data[:-1]).nonzero()[0]
        raise ValueError('time-coordinate not increasing at {}'.format(indx))

    # determine delta_time with greatest common divisor (numpy v1.15+)
    if delta_time is None:
        delta_time = np.gcd.reduce(np.diff(time_data))

    time_dim = 1 + (time_data.max() - time_data.min()) // delta_time
    if time_dim == time_data.size:
        return msm

    time_data_full = np.arange(time_data.min(),
                               time_data.max() + delta_time,
                               delta_time)
    mask = np.in1d(time_data_full, time_data)
    if time_axis == 0:
        data_full = np.empty((time_dim, msm.values.shape[1]),
                             dtype=msm.values.dtype)
        data_full[mask, :] = msm.values
        for ii in (~mask).nonzero()[0]:
            data_full[ii, :] = data_full[ii-1, :]
    else:
        data_full = np.empty((msm.values.shape[0], time_dim),
                             dtype=msm.values.dtype)
        data_full[:, mask] = msm.values
        for ii in (~mask).nonzero()[0]:
            data_full[ii, :] = data_full[ii-1, :]

    msm.values = data_full
    return msm


def get_fig_coords(data_in) -> dict:
    """
    get labels of the X/Y axis

    Returns
    -------
    dict: {'X': {'label': value, 'data': value},
           'Y': {'label': value, 'data': value}}
    """
    if isinstance(data_in, np.ndarray):
        ylabel, xlabel = ['row', 'column']
        ydata = np.arange(data_in.shape[0])
        xdata = np.arange(data_in.shape[1])
    else:
        ylabel, xlabel = data_in.dims
        ydata = data_in.coords[ylabel].data
        xdata = data_in.coords[xlabel].data

    return {'X': {'label': xlabel, 'data': xdata},
            'Y': {'label': ylabel, 'data': ydata}}


def get_xdata(xdata, use_steps: bool) -> tuple:
    """
    The X-coordinate from the data in object msm is checked for data gaps.
    The data of the X-coordinate is extended to avoid interpolation over
    missing data.

    A list of indices to all data gaps is also returned which can be used to
    update the Y-coordinate.
    """
    # this function only works when the X-coordinate is of type integer and
    # increasing
    if not np.issubdtype(xdata.dtype, np.integer):
        raise ValueError('x-coordinate not of type integer')
    if not np.all(xdata[1:] > xdata[:-1]):
        indx = (xdata[1:] <= xdata[:-1]).nonzero()[0]
        raise ValueError('x-coordinate not increasing at {}'.format(indx))

    xstep = np.gcd.reduce(np.diff(xdata))
    gap_list = 1 + (np.diff(xdata) > xstep).nonzero()[0]
    for indx in reversed(gap_list):
        xdata = np.insert(xdata, indx, xdata[indx])
        xdata = np.insert(xdata, indx, xdata[indx-1] + xstep)
        xdata = np.insert(xdata, indx, xdata[indx-1] + xstep)

    if use_steps:
        xdata = np.append(xdata, xdata[-1] + xstep)

    return (xdata, gap_list)



"""
This file is part of pyS5p

https://github.com/rmvanhees/pys5p.git

Implementation of the Relative Least-Squares regression (RLS).

The RLS regression is used to find the linear dependence y(x) = c0 + c1 * x
that best describes the data before and after correction, using absolute
residuals y_i - (c0 + c1 * x_i) divided by the expected signals c1 * x_i in
the least-squares sum. Offset c0 has an arbitrary size and should not affect
the fit result. Weight factors are determined such to effectively spread the
data points evenly over the whole range of x, making the result less
sensitive to the actual spacing between the data points.

Copyright (c) 2019-2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import numpy as np


# pylint: disable=invalid-unary-operand-type
# - local functions --------------------------------
def __rls_fit_all(xx, yy):
    """
    RLS regression without masked pixels
    """
    wght = np.concatenate(([2 * (xx[1] - xx[0])],
                           xx[2:] - xx[0:-2],
                           [2 * (xx[-1] - xx[-2])]))
    wx1 = wght / xx
    wx2 = wght / xx ** 2

    # calculate the Q elements
    q00 = np.sum(wght)
    q01 = np.sum(wx1)
    q02 = np.sum(wx2)

    q11 = yy @ wx1
    q12 = yy @ wx2
    q22 = yy ** 2 @ wx2

    # calculate the Z elements
    zz1 = q00 * q02 - q01 ** 2
    zz2 = q00 * q12 - q01 * q11
    zz3 = q02 * q11 - q01 * q12

    cc0 = zz2 / zz1
    cc1 = zz3 / zz1

    if xx.size == 2:
        chi2 = np.zeros(cc0.shape[0], dtype=float)
        sc0 = np.zeros(cc0.shape[0], dtype=float)
        sc1 = np.zeros(cc1.shape[0], dtype=float)
    else:
        chi2 = np.abs(q22 - q12 * cc0 - q11 * cc1) / (xx.size - 2)
        sc0 = np.sqrt(q00 * chi2 / zz1)
        sc1 = np.sqrt(q02 * chi2 / zz1)

    return (cc0, cc1, sc0, sc1, chi2)


def __rls_fit_masked(xx, yy, mask, num):
    """
    RLS regression with masked pixels
    """
    # generate weight factor per pixel
    wght = np.empty(yy.shape, dtype=float)
    wght[:] = np.concatenate(([2 * (xx[1] - xx[0])],
                              xx[2:] - xx[0:-2],
                              [2 * (xx[-1] - xx[-2])]))
    wght[~mask] = 0.
    wx1 = wght / xx
    wx2 = wght / xx ** 2

    # calculate the Q elements
    q00 = np.sum(wght, axis=1)
    q01 = np.sum(wx1, axis=1)
    q02 = np.sum(wx2, axis=1)

    q11 = np.sum(wx1 * yy, axis=1)
    q12 = np.sum(wx2 * yy, axis=1)
    q22 = np.sum(wx2 * yy ** 2, axis=1)

    # calculate the Z elements
    zz1 = q00 * q02 - q01 ** 2
    zz2 = q00 * q12 - q01 * q11
    zz3 = q02 * q11 - q01 * q12

    zz1[zz1 == 0] = float.fromhex('0x1.ep-122')
    cc0 = zz2 / zz1
    cc0[num < 2] = np.nan
    cc1 = zz3 / zz1
    cc1[num < 2] = np.nan

    chi2 = np.abs(q22 - q12 * cc0 - q11 * cc1) / np.clip(num - 2, 1, None)
    chi2[num <= 2] = 0
    sc0 = np.sqrt(q00 * chi2 / zz1)
    sc1 = np.sqrt(q02 * chi2 / zz1)

    return (cc0, cc1, sc0, sc1, chi2)


# - main functions _--------------------------------
def rls_fit(xx, yy, samples_not_saturated=None):
    """
    Fast implementation of the RLS regression finding linear dependence
      y(x) = c0 + c1 * x

    Parameters
    ----------
    xx :  ndarray, shape (M,)
       X-coordinates of the M sample points (xx[i], yy[..., i])
       The array values have to be positive and increasing
    yy :  ndarray, shape (..., M)
       Y-coordinates of the sample points
    samples_not_saturated : ndarray, optional
       number of sample points to include in the fit, startinge from the first

    Returns
    -------
    c0, c1, std_c0, std_c1, chi2 : tuple of ndarrays
       coefficients of the linear dependence and their standard deviations

    Raises
    ------
    RuntimeError
       if M < 2: too few points for a fit
       if M_xx != M_yy: number of samples not equal for xx, yy
       if yy.shape[:-1] != samples_not_saturated.shape:
          arrays yy and mask do not have equal shapes

    Notes
    -----
    The standard deviations can only be calculated when the number of samples
    are larger than two, else the standard deviations are equal to zero.

    The coefficients are set to NaN when the number of samples are less than 2.

    Examples
    --------
    How to use this function:
    >>> dimx = dimy = 2048
    >>> texp = np.array([1e-3, 2e-3, 2.5e-3, 5e-3, 1e-2, 5e-2, 1e-1, .5, 1, 5])
    >>> alpha = np.random.random_sample((dimy, dimx)) + 11
    >>> beta = np.random.random_sample((dimy, dimx)) + 3
    >>> data = alpha[..., np.newaxis] + texp * beta[..., np.newaxis]
    >>> cc0, cc1, sc0, sc1, chi2 = rls_fit(texp, data)
    >>> mask_nan = np.isfinite(cc0)
    >>> print('cc0 ', np.nanmean(cc0),
    ...       np.abs(cc0[mask_nan] - alpha[mask_nan]).max())
    cc0  11.49983030605162 7.105427357601002e-15
    >>> print('sc0 ', np.nanmean(sc0))
    sc0  5.0509648555957586e-08
    """
    if xx.size < 2:
        raise RuntimeError('too few sample points for a fit')
    if xx.size != yy.shape[-1]:
        raise RuntimeError('number of samples not equal for xx, yy')

    # perform all computations on 2 dimensional arrays
    data = yy.reshape(-1, xx.size)

    # ---------- include all data samples ----------
    if samples_not_saturated is None:
        cc0, cc1, sc0, sc1, chi2 = __rls_fit_all(xx, data)

    # ---------- exclude several of the last data samples ----------
    else:
        if yy.shape[:-1] != samples_not_saturated.shape:
            raise RuntimeError('arrays yy and mask do not have equal shapes')
        buff = (np.arange(yy.size) % xx.size).reshape(yy.shape)
        mask = buff <= samples_not_saturated[:, :, np.newaxis]
        del buff
        if np.isnan(yy).any():
            mask[np.isnan(yy)] = False
        mask = mask.reshape(-1, xx.size)

        cc0, cc1, sc0, sc1, chi2 = __rls_fit_masked(
            xx, data, mask, samples_not_saturated.reshape(-1))

    return (cc0.reshape(yy.shape[:-1]), cc1.reshape(yy.shape[:-1]),
            sc0.reshape(yy.shape[:-1]), sc1.reshape(yy.shape[:-1]),
            chi2.reshape(yy.shape[:-1]))


# --------------------------------------------------
def rls_fit0(xx, yy, samples_not_saturated=None):
    """
    Fast implementation of RLS regression finding linear dependence
      y(x) = c1 * x

    Parameters
    ----------
    xx :  ndarray, shape (M,)
       X-coordinates of the M sample points (xx[i], yy[..., i])
       The array values have to be positive and increasing
    yy :  ndarray, shape (..., M)
       Y-coordinates of the sample points
    samples_not_saturated : ndarray, optional
       number of sample points to include in the fit, startinge from the first

    Returns
    -------
    c1, std_c1 : tuple of ndarrays
       coefficient of the linear dependence and its standard deviation

    Raises
    ------
    RuntimeError
       if M < 2: too few points for a fit
       if M_xx != M_yy: number of samples not equal for xx, yy
       if yy.shape != mask.shape: arrays yy and mask do not have equal shapes

    See also
    --------
       pys5p.rls.rls_fit
    """
    if xx.size < 2:
        raise RuntimeError('too few points for a fit')
    if xx.size != yy.shape[-1]:
        raise RuntimeError('number of samples not equal for xx, yy')

    # perform all computations on 2 dimensional arrays
    data = yy.reshape(-1, xx.size)

    # ---------- include all data samples ----------
    if samples_not_saturated is None:
        wght = np.concatenate(([2 * (xx[1] - xx[0])],
                               xx[2:] - xx[0:-2],
                               [2 * (xx[-1] - xx[-2])]))
        wx1 = wght / xx
        wx2 = wght / xx ** 2

        # calculate the Q elements
        q00 = np.sum(wght)
        q11 = data @ wx1
        q22 = data ** 2 @ wx2

        # calculate fit parameter and its variance
        cc1 = q11 / q00
        chi2 = np.abs(q22 - q00 * cc1 ** 2) / (xx.size - 1)
        sc1 = np.sqrt(chi2 / q00)

    # ---------- exclude several of the last data samples ----------
    else:
        if yy.shape[:-1] != samples_not_saturated.shape:
            raise RuntimeError('arrays yy and mask do not have equal shapes')
        buff = (np.arange(yy.size) % xx.size).reshape(yy.shape)
        mask = buff <= samples_not_saturated[:, :, np.newaxis]
        del buff
        mask = mask.reshape(-1, xx.size)
        num = samples_not_saturated.reshape(-1)

        # generate weight factor per pixel
        wght = np.empty(data.shape, dtype=float)
        wght[:] = np.concatenate(([2 * (xx[1] - xx[0])],
                                  xx[2:] - xx[0:-2],
                                  [2 * (xx[-1] - xx[-2])]))
        wght[~mask] = 0.
        wx1 = wght / xx
        wx2 = wght / xx ** 2

        # calculate the Q elements
        q00 = np.sum(wght, axis=1)
        q11 = np.sum(wx1 * data, axis=1)
        q22 = np.sum(wx2 * data ** 2, axis=1)

        # calculate fit parameter and its variance
        cc1 = q11 / q00
        cc1[num < 2] = np.nan
        chi2 = np.abs(q22 - q00 * cc1 ** 2) / np.clip(num - 1, 1, None)
        chi2[num <= 2] = 0
        sc1 = np.sqrt(chi2 / q00)

    return (cc1.reshape(yy.shape[:-1]),
            sc1.reshape(yy.shape[:-1]),
            chi2.reshape(yy.shape[:-1]))



"""
This file is part of pyS5p

https://github.com/rmvanhees/pys5p.git

Implements a lite interface with the xarray::DataArray

Copyright (c) 2020 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from pathlib import PurePath

import numpy as np
import xarray as xr


def __get_attrs(dset, field: str) -> dict:
    """
    Return attributes of the HDF5 dataset

    Parameters
    ----------
    dset :  h5py.Dataset
       h5py dataset from which the attributes are read
    field : str
       Name of field in compound dataset

    Returns
    -------
    dict with numpy arrays
    """
    _field = None
    if field is not None:
        try:
            _field = {'name': field,
                      'oneof': len(dset.dtype.names),
                      'index': dset.dtype.names.index(field)}
        except Exception as exc:
            raise RuntimeError('field {} not found in dataset {}'.format(
                field, dset.name)) from exc
        # print('_field ', _field)

    attrs = {}
    for key in dset.attrs:
        if key in ('CLASS', 'DIMENSION_LIST', 'NAME', 'REFERENCE_LIST',
                   '_Netcdf4Dimid'):
            continue

        attr_value = dset.attrs[key]
        # print('# ----- ', key, type(attr_value), attr_value)
        if isinstance(attr_value, np.ndarray):
            if len(attr_value) == 1:
                attr_value = attr_value[0]
                # print('# ----- ', key, type(attr_value), attr_value)
            elif _field is not None:
                if len(attr_value) == _field['oneof']:
                    attr_value = attr_value[_field['index']]
                # elif isinstance(attr_value, np.void):
                #    attr_value = attr_value[0]

        attrs[key] = (attr_value.decode('ascii')
                      if isinstance(attr_value, bytes) else attr_value)

    return attrs


def __get_coords(dset, data_sel: tuple) -> list:
    """
    Return coordinates of the HDF5 dataset

    Parameters
    ----------
    dset :  h5py.Dataset
       h5py dataset from which the data is read
    data_sel :  tuple of slice(s)
       A numpy slice generated for example numpy.s_
       Default read while array

    Returns
    -------
    A sequence of tuples [(dims, data), ...]
    """
    coords = []
    if len(dset.dims) == dset.ndim:
        try:
            for ii, dim in enumerate(dset.dims):
                name = PurePath(dim[0].name).name
                if name.startswith('row') or name.startswith('column'):
                    name = name.split(' ')[0]

                buff = dim[0][()]
                if np.all(buff == 0):
                    buff = np.arange(dim[0].size, dtype=dim[0].dtype)

                coords.append((name, buff
                               if data_sel is None else buff[data_sel[ii]]))
        except RuntimeError:
            coords = []

    return coords


def __set_coords(dset, data_sel: tuple, dims: list) -> list:
    """
    Set coordinates of the HDF5 dataset

    Parameters
    ----------
    dset :  h5py.Dataset or numpy.array
       h5py dataset from which the data is read
    data_sel :  tuple of slice(s)
       A numpy slice generated for example numpy.s_
       Default read while array
    dims : list of strings
       Alternative names for the dataset dimensions if not attached to dataset
       Default coordinate names are ['time', ['row', ['column']]]

    Returns
    -------
    A sequence of tuples [(dims, data), ...]
    """
    if dims is None:
        if dset.ndim > 3:
            raise ValueError('not implemented for ndim > 3')

        dims = ['time', 'row', 'column'][-dset.ndim:]

    coords = []
    for ii in range(dset.ndim):
        buff = np.arange(dset.shape[ii], dtype='u4')
        if data_sel is not None:
            buff = buff[data_sel[ii]]
        coords.append((dims[ii], buff))

    return coords


def __get_data(dset, data_sel: tuple, field: str):
    """
    Return data of the HDF5 dataset

    Parameters
    ----------
    dset :  h5py.Dataset
       h5py dataset from which the data is read
    data_sel :  numpy slice
       A numpy slice generated for example numpy.s_
       Default read while array
    field : str
       Name of field in compound dataset or None

    Returns
    -------
    Numpy array

    Notes
    -----
    Read floats always as doubles
    """
    if data_sel is None:
        data_sel = ()

    if np.issubdtype(dset.dtype, np.floating):
        data = dset.astype(float)[data_sel]
        return data

    if field is None:
        return dset[data_sel]

    data = dset.fields(field)[data_sel]
    if np.issubdtype(data.dtype, np.floating):
        data = data.astype(float)
    return data


def h5_to_xr(h5_dset, field=None, data_sel=None, dims=None):
    """
    Create xarray::DataArray from a HDF5 dataset (with dimension scales)

    Parameters
    ----------
    h5_dset :  h5py.Dataset
       Data, dimensions, coordinates and attributes are read for this dataset
    field : str
       Name of field in compound dataset or None
    data_sel :  numpy slice
       A numpy slice generated for example numpy.s_
    dims :  list of strings
       Alternative names for the dataset dimensions if not attached to dataset

    Returns
    -------
    xarray.DataArray

    Notes
    -----
    If data_sel is used to select data from a dataset then the number of
    dimensions of data_sel should agree with the HDF5 dataset. Thus allowed
    values for data_sel are:
    * [always]: (), np.s_[:], np.s_[...]
    * [1-D dataset]: np.s_[:-1], np.s_[0]
    * [2-D dataset]: np.s_[:-1, :], np.s_[0, :], np.s_[:-1, 0]
    * [3-D dataset]: np.s_[:-1, :, 2:4], np.s_[0, :, :], np.s_[:-1, 0, 2:4]
    But not np.s_[0, ...], np.s_[..., 4]

    Examples
    --------
    >>> fid = h5py.File(flname, 'r')        # flname is a HDF5/netCDF4 file
    >>> xdata = h5_to_xr(fid['signal'])
    >>> fid.close()

    Combine Tropomi SWIR data of band7 and band8
    >>> fid = h5py.File(s5p_b7_prod, 'r')   # Tropomi band7 product
    >>> xdata7 = h5_to_xr(fid['signal'])
    >>> fid.close()
    >>> fid = h5py.File(s5p_b8_prod, 'r')   # Tropomi band8 product
    >>> xdata8 = h5_to_xr(fid['signal'])
    >>> fid.close()
    >>> xdata = xr.concat((xdata7, xdata8), dim='spectral_channel')

    Optionally, fix the 'column' dimension
    >>> xdata = xdata.assign_coords(
    >>> ... column=np.arange(xdata.column.size, dtype='u4'))
    """
    if data_sel is not None:
        if data_sel in (np.s_[:], np.s_[...], np.s_[()]):
            data_sel = None

    # Name of this array
    name = PurePath(h5_dset.name).name

    # Values for this array
    data = __get_data(h5_dset, data_sel, field)

    # Coordinates (tick labels) to use for indexing along each dimension
    coords = []
    if dims is None:
        coords = __get_coords(h5_dset, data_sel)
    if not coords:
        coords = __set_coords(h5_dset, data_sel, dims)

    # Attributes to assign to the array
    attrs = __get_attrs(h5_dset, field)

    # check if dimension of dataset and coordinates agree
    if data.ndim < len(coords):
        for ii in reversed(range(len(coords))):
            if np.isscalar(coords[ii][1]):
                del coords[ii]

    return xr.DataArray(data, name=name, attrs=attrs, coords=coords)


def data_to_xr(data, dims=None, name=None, long_name=None, units=None):
    """
    Create xarray::DataArray from a dataset

    Parameters
    ----------
    data :  array-like
       Data to be stored in the xarray
    dims :  list of strings
       Names for the dataset dimensions
    name : str
       A string that names the instance
    units :  str
       Units of the data, default: '1'
    long_name : str
       Long name describing the data, default: empty string

    Returns
    -------
    xarray.DataArray
    """
    coords = __set_coords(data, None, dims)
    attrs = {}
    attrs['units'] = '1' if units is None else units
    attrs['long_name'] = '' if long_name is None else long_name

    return xr.DataArray(data, name=name, attrs=attrs, coords=coords)
